package common;

import java.util.ArrayList;



public class Job {
	public static final int STATUS_RUNNING=0;
	public static final int STATUS_COMPLETE=1;
	public static final int STATUS_WAITING=2;
	public static final int STATUS_FAILED=3;
	public static final int STATUS_ABORTED=4;
	
	public static final int KB=1024;
	
	public static final String STATUS_NAMES[]={"Running", "Complete", "Waiting", "Failed", "Aborted"};
	
	protected String description;
	private int size;
	private int position;
	protected ArrayList<JobListenerI> listeners=new ArrayList<JobListenerI>(2);
	protected int status;
	protected long startTime;
	protected long endTime;
	
	public Job() {
		reset();
	}
	
	public Job(JobListenerI listener) {
		reset();
		addJobListener(listener);
	}
	
	public void reset() {
		startTime=System.currentTimeMillis();
		endTime=0;
		status=STATUS_WAITING;
		position=0;
	}
	
	public int getStatus() {
		return status;
	}
	public int  setStatus(int status) {
		this.status = status;
		return notifyListeners();
	}
	
	public void done() {
		endTime=System.currentTimeMillis();
		setStatus(STATUS_COMPLETE);
	}
	public int notifyListeners() {
		for (int i=0;i<listeners.size();i++) {
			JobListenerI listener=(JobListenerI)listeners.get(i);
			int rc=listener.jobStatus(this);
			if (rc!=StatusI.CONTINUE) {
				status=STATUS_ABORTED;
			}
		}
		return status;
	}
	
	public long getStartTime() {
		return startTime;
	}

	public String getDescription() {
		return description;
	}

	public void setDescription(String description) {
		this.description = description;
	}

	public int getSize() {
		return size;
	}

	public void setSize(int size) {
		this.size = size;
	}

	public int getPosition() {
		return position;
	}

	public int setPosition(int position) {
		this.position = position;
		return notifyListeners();
	}
	public void addJobListener(JobListenerI listener) {
		if (listener != null && !listeners.contains(listener)) {
			listeners.add(listener);
		}
	}
}
