package common.ts;

import java.io.FileInputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import common.StreamDescriptor;
import common.StreamIdentifier;

import demux.ts.TsRawPacketPart;

public class ProgramMapTable extends TsTable {
	private static final StreamIdentifier streamIdentifier=new StreamIdentifier();
	private int pcrPid;
	//private short programNumber;
	//private TreeMap streamDescriptorMap;
	private Map<Integer, StreamDescriptor> selectedStreamDescriptorMap;
	private ProgramInfo programInfo;
	
	public ProgramMapTable(int pid, short programNumber) {
		super(pid, TsConstants.PROGRAM_MAP_TABLE_ID);
		programInfo=new ProgramInfo(programNumber, pid);
		//this.programNumber=programNumber;
	}
	public ProgramMapTable(ProgramInfo programInfo) {
		super(programInfo.getPmtPid(), TsConstants.PROGRAM_MAP_TABLE_ID);
		this.programInfo=programInfo;
	}
	/**
	 * @param streamDescriptorMap
	 * @param reparse reparse the stream in buffer, or just use the last parsed stream
	 * @return
	 */
	public TsRawPacketPart[] filter(boolean reparse) {
		if (selectedStreamDescriptorMap==null) {
			return buffer;
		}
		if (reparse || out==null) {
			ByteBuffer tmp=mergeIn(buffer);
			out=ByteBuffer.allocate(1021);
			copy(tmp, out, 9);
			/*programNumber=*/tmp.getShort(4);
			short pcrPid=tmp.getShort();
			this.pcrPid=pcrPid & 0x1fff;
			out.putShort(pcrPid);
			short piLen=tmp.getShort();
			out.putShort(piLen);
			copy(tmp,out,piLen & 0xfff);
			while (tmp.remaining()>4) {
				/*byte type=*/tmp.get();
				short pid=tmp.getShort();
				short esLen=tmp.getShort();
				int esLenI=esLen & 0xfff;
				StreamDescriptor sd=selectedStreamDescriptorMap.get(Integer.valueOf(pid&0x1fff));
				if (sd != null) {
					//Set type based on stream descriptor in case stream has been cored
					out.put(sd.getTsType());
					out.putShort(pid);
					out.putShort(esLen);
					copy(tmp, out, esLenI);
				} else {
					tmp.position(tmp.position()+(esLenI));
				}
			}
			finalizeOut();
		}
		return mergeOut(buffer);
	}

	public TsRawPacketPart[] parse() {
		//int len=(pmtBuffer[0].getByteBuffer().getShort(pmtBuffer[0].getTsSyncOffset()+6) & 0xfff)+4;
		ByteBuffer tmp=mergeIn(buffer);
		/*tmp.position(11);
		/short pcrPid=tmp.getShort();*/
		tmp.position(9);
		pcrPid=tmp.getShort() & 0x1fff;
		
		int programInfoLength=tmp.getShort() & 0xfff;
 		tmp.position(tmp.position()+programInfoLength);
		Map<Integer, StreamDescriptor> streamDescriptorMap=new TreeMap<Integer, StreamDescriptor>();
		while (tmp.remaining()>4) {
			byte type=tmp.get();
			int pid=tmp.getShort() & 0x1fff;
			//short esLen=tmp.getShort();
			int esInfoLength=tmp.getShort() & 0xfff;
			StreamDescriptor descriptor=streamIdentifier.getTsStreamDescriptor(pid, type);
			descriptor.setProgramNumber(programInfo.getProgramNumber());
			streamDescriptorMap.put(Integer.valueOf(pid), descriptor);
			tmp.position(tmp.position()+esInfoLength);
		}
		programInfo.setStreamDescriptorMap(streamDescriptorMap);
		found=true;
		notifyListeners();
		return buffer;
	}
	public int getPcrPid() {
		return pcrPid;
	}
	public void setPcrPid(int pcrPid) {
		this.pcrPid = pcrPid;
	}
/*	public short getProgramNumber() {
		return programNumber;
	}
	public void setProgramNumber(short programNumber) {
		this.programNumber = programNumber;
	}*/
	
	public static void main(String args[]) {
		try {
			FileChannel in=new FileInputStream("D:/Video/Sesame Street_KRMADT_2009_11_30_08_58_00_new.ts").getChannel();
			ByteBuffer buffer=in.map(FileChannel.MapMode.READ_ONLY,188,188);
			ProgramMapTable pmt=new ProgramMapTable(0,(short)1);
			
			pmt.append(new TsRawPacketPart(buffer));
			pmt.parse();
			//Map streamMap=pmt.getStreamDescriptorMap();
			//System.out.println(streamMap);
			System.out.println(pmt.getProgramInfo());

			/*TsRawPacketPart out[]=pmt.filter(streamMap, true);
			ByteBuffer test=out[0].getByteBuffer();
			for (int i=0;i<buffer.capacity();i++) {
				if (buffer.get(i) != test.get(i)) {
					System.out.println("Failed at: " + i + " - " + Integer.toString(buffer.get(i) & 0xff, 16) + ':' + Integer.toString(test.get(i) & 0xff, 16));
				}
			}*/
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/*public TreeMap getStreamDescriptorMap() {
		return streamDescriptorMap;
	}*/

	public void setSelectedStreamDescriptorMap(Map<Integer, StreamDescriptor> selectedStreamDescriptorMap) {
		this.selectedStreamDescriptorMap = selectedStreamDescriptorMap;
	}
	public Map<Integer, StreamDescriptor> getSelectedStreamDescriptorMap() {
		return this.selectedStreamDescriptorMap;
	}
	public ProgramInfo getProgramInfo() {
		return programInfo;
	}
	public void setSelectedProgramNumbers(short[] selectedProgramNumbers, TableHandler tableHandler) {
		for (int i=0;i<selectedProgramNumbers.length;i++) {
			if (programInfo.getProgramNumber()==selectedProgramNumbers[i]) {
				return;
			}
		}
		tableHandler.remove(this);
	}
	public void setStreamList(List<StreamDescriptor> list) {
		HashMap<Integer, StreamDescriptor> map=new HashMap<Integer, StreamDescriptor>();
		for (int i=0;i<list.size();i++) {
			StreamDescriptor sd=list.get(i);
			map.put(Integer.valueOf(sd.getStreamId()), sd);
		}
		this.setSelectedStreamDescriptorMap(map);
	};
	/**
	 * @param buffer
	 * @param offset 8 for m2ts, 4 for ts
	 * @return array of programNumber and PMT PIDs
	 */
	public static int[][] getPmtPids(final ByteBuffer buffer, final int offset) {
		int len=((buffer.get(offset+2) & 0x0F) << 8) + buffer.get(offset+3) & 0xFF;
		int programs=(len-9)/4;
		int tmpPids[][]=new int[programs][];
		int j=0;
		for (int i=0;i<programs;i++) {
			int programId=buffer.getShort(offset + 9 + i*4) & 0xffff;
			//ProgramId=0 means network PID
			if (programId!=0) {
				tmpPids[j++]=new int[]{programId, buffer.getShort(offset + 11 + i*4) & 0x1FFF};
			}
		}
		if (j!=tmpPids.length) {
			int pmtPids[][]=new int[j][2];
			for (int i=0;i<j;i++) {
				pmtPids[i]=tmpPids[i];
			}
			return pmtPids;
		} else {
			return tmpPids;
		}
	}	
}
