package demux.ts;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import muxer.MuxManager;
import muxer.ts.TsRawMuxer;

import common.FileStream;
import common.Job;
import common.JobStreamI;
import common.StreamDescriptor;
import common.bluray.PlayList;
import common.bluray.PlayListAnalyzer;
import common.bluray.PlayList.Branch;

public class M2tsRawDemuxer extends TsRawDemuxer {
	public int getOffset() {
		return 4;
	}
	protected int getPacketSize() {
		return M2TS_PACKET_SIZE;
	}
	protected TsRawPacketPart getPacketPart(JobStreamI stream) throws IOException{
		return new M2tsRawPacketPart(stream.slice(M2TS_PACKET_SIZE),4);
	}
	public Map<Integer, StreamDescriptor> getStreamDescriptorMap(JobStreamI stream, MuxManager muxManager) throws IOException {
		Map<Integer, StreamDescriptor> map=super.getStreamDescriptorMap(stream, muxManager);
		
		if (stream instanceof FileStream) {
			FileStream fileStream=(FileStream)stream;
			File file=fileStream.getFiles()[0];
			File playListDir=new File(file.getParent(), "../PLAYLIST");
			
			if (playListDir.isDirectory()) {
				PlayListAnalyzer analyzer=new PlayListAnalyzer();
				File playListFileArray[]=playListDir.listFiles();
				String fileName=file.getName().toUpperCase();
				
				for (int i=0;i<playListFileArray.length;i++) {
					PlayList playList=new PlayList();
					playList.parse(playListFileArray[i]);
					List<Branch> branchList=playList.getBranchList();
					
					for (int j=0;j<branchList.size();j++) {
						if (fileName.equals(branchList.get(j).getFileName())) {
							Map<Integer, StreamDescriptor> playListMap=branchList.get(j).getStreamDescriptorMap();
							for (Iterator<Integer> ki=map.keySet().iterator();ki.hasNext();) {
								Integer key=ki.next();
								StreamDescriptor in=map.get(key);
								
								if (in.getLanguage()==null) {
									StreamDescriptor listSD=playListMap.get(key);
									if (listSD != null) {
										in.setLanguage(listSD.getLanguage());
									}
								}
							}
						}
					}
				}
			}
		}
		return map;
	}
	public static void main(String args[]) {
		String titleName="J:/BDMV/STREAM/00008.m2ts";
		
		try {
			M2tsRawDemuxer demuxer=new M2tsRawDemuxer();
			MuxManager muxManager=new MuxManager(demuxer);
			FileStream job=new FileStream(new Job(null), new File[]{new File(titleName)});
			Map map=muxManager.getStreamDescriptorMap(job);
			ArrayList streamList=new ArrayList(3);
			streamList.add(map.get(Integer.valueOf(0x1011)));
			streamList.add(map.get(Integer.valueOf(0x1100)));
			//streamList.add(map.get(Integer.valueOf(0x1200)));
			demuxer.setStreams(streamList);
			System.out.println(map);
			//muxManager.add(new M2tsRawMuxer(new File("F:/Black Hawk Down.m2ts")));
			muxManager.add(new TsRawMuxer(new FileOutputStream("F:/T2.ts").getChannel()));
			muxManager.mux(job);
			//PacketBuffer packetBuffer=muxManager.getPacketBuffer();
			//packetBuffer.setRegulator(PacketBuffer.ALWAYS_FULL);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}	
}
