package muxer.filter;

import java.io.IOException;
import java.nio.ByteBuffer;

import muxer.MuxFilter;

import common.packet.Packet;

import demux.ts.TsInfo;
import demux.ts.TsRawPacketPart;

public class AvcSequenceEndMuxFilter extends MuxFilter {
	private static final int SEQUENCE_END=0x0000010a;
	private int streamId;
	private int lastPacketBytes=0xffffffff;

	public AvcSequenceEndMuxFilter(int streamId) {
		this.streamId=streamId;
	}
	
	public int addPacket(Packet packet) throws IOException {
		if (packet.getStreamId()==streamId) {
			TsRawPacketPart packetPart=(TsRawPacketPart)packet.getPayload();
			TsInfo info=packetPart.getTsInfo();
			ByteBuffer buffer=packetPart.getByteBuffer();
			int offset=packetPart.getTsSyncOffset()+4;
			if (info.adaptationControl >= 0x20) {
				offset += (packetPart.get(offset) & 0xff) + 1; 
			}
			final int myLimit=buffer.limit();
			for (int i=0;i + offset < myLimit;i++) {
				lastPacketBytes=lastPacketBytes<<8 | (buffer.get(offset + i) & 0xff);
				if (lastPacketBytes==SEQUENCE_END) {
					if (buffer.isReadOnly()) {
						buffer=packetPart.setReadWrite();
					}
					buffer.put(offset+i, (byte)0x1a);
				}
			}
		} 
		return super.addPacket(packet);
	}
	
	public int getPosition() {
		return POSITION_PRE_BUFFER;
	}	
}
