package ui;

import io.NameTypeFileFilter;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.List;

import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.ListSelectionModel;
import javax.swing.UIManager;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.text.BadLocationException;

import ui.tonmt.ToNmtPreferenceDialog;
import ui.tonmt.ToNMTPrefs;
import util.Eac3toTsMuxerConverter;
import util.InternalStreamConverter;
import util.StreamConverter;

import common.DiskFinder;
import common.DiskStream;
import common.FileStream;
import common.Job;
import common.Logger;
import common.LoggerAppenderI;
import common.StreamDescriptor;
import common.VideoStreamDescriptor;
import common.bluray.PlayListAnalyzer;

import demux.pes.XplHelper;

public abstract class ToUI extends JFrame {
	private static final String LIST_BOX_PROTOTYPE="ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890";
	private static final String SAVE_AS_TEXT="Save As...";

//	private static final String CANCEL_TEXT="Cancel";
	
	protected JFileChooser inputFileChooser;
	protected JFileChooser saveAsFileChooser;
	protected DiskFinder diskFinder;
	private Logger logger=Logger.getLogger(this.getClass());
	private StreamConverter converter;
	private StreamDescriptor sd[][];
	
	protected abstract ToNmtPreferenceDialog getPreferenceDialog();
	
	protected abstract ToNMTPrefs getPrefs();
	/*public static void main(String args[]) {
		ToUI frame = new ToUI();
		frame.setVisible(true);
	}*/
	protected ToUI(String name) {
		setName(name);
		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e) {
			e.printStackTrace();
			return;
		} 
		
		setTitle(getName());
		getContentPane().setLayout(new GridBagLayout());
		setBounds(100, 100, 600, 550);
		Insets insets=new Insets(1,1,1,1);
		
		int y=0;
		final JScrollPane inputFileNameScrollPane = new JScrollPane();
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.anchor=GridBagConstraints.NORTH;
		gbc.fill=GridBagConstraints.BOTH;
		gbc.gridwidth=2;
		gbc.weightx=1;
		gbc.gridheight=3;
		gbc.insets=insets;
		getContentPane().add(inputFileNameScrollPane, gbc);
		
		final JTextArea inputFileNameTextArea = new JTextArea();
		inputFileNameTextArea.setEditable(false);
		inputFileNameTextArea.setRows(4);
		inputFileNameScrollPane.setViewportView(inputFileNameTextArea);
		
		final JButton autoButton=new JButton("Auto Source");
		gbc = new GridBagConstraints();
		gbc.anchor=GridBagConstraints.NORTH;
		gbc.fill=GridBagConstraints.HORIZONTAL;
		gbc.gridx=2;
		gbc.insets=insets;
		getContentPane().add(autoButton, gbc);
		
		final JButton sourceFileNameButton = new JButton();
		gbc = new GridBagConstraints();
		gbc.gridy=++y;
		gbc.fill=GridBagConstraints.HORIZONTAL;
		gbc.gridx=2;
		gbc.insets=insets;
		getContentPane().add(sourceFileNameButton, gbc);
		sourceFileNameButton.setText("Source...");
		
		final JButton preferencesButton = new JButton();
		gbc = new GridBagConstraints();
		gbc.gridy=++y;
		gbc.fill=GridBagConstraints.HORIZONTAL;
		gbc.gridx=2;
		gbc.insets=insets;
		getContentPane().add(preferencesButton, gbc);
		preferencesButton.setText("Preferences");

		//Audio
		final JLabel audioLabel=new JLabel("Audio:");
		gbc = new GridBagConstraints();
		gbc.gridy=++y;
		gbc.anchor=GridBagConstraints.WEST;
		gbc.insets=insets;
		getContentPane().add(audioLabel,gbc);
		
		final JScrollPane audioListScrollPane = new JScrollPane();
		gbc = new GridBagConstraints();
		gbc.gridy=++y;
		//gbc.weighty=.2;
		gbc.weightx=1;
		gbc.fill=GridBagConstraints.HORIZONTAL;
		gbc.gridwidth=3;
		gbc.insets=insets;
		getContentPane().add(audioListScrollPane, gbc);
		
		final JList audioListBox = new JList();
		final DefaultListModel audioListModel=new DefaultListModel();
		audioListBox.setModel(audioListModel);
		audioListBox.setPrototypeCellValue(LIST_BOX_PROTOTYPE);
		audioListBox.setVisibleRowCount(4);
		audioListBox.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		audioListScrollPane.setViewportView(audioListBox);
		
		//Subtitles
		final JLabel subtitleLabel=new JLabel("Subtitle:");
		gbc = new GridBagConstraints();
		gbc.gridy=++y;
		gbc.anchor=GridBagConstraints.WEST;
		gbc.insets=insets;
		getContentPane().add(subtitleLabel,gbc);
		
		final JScrollPane subtitleListScrollPane = new JScrollPane();
		gbc = new GridBagConstraints();
		gbc.gridy=++y;
		//gbc.weighty=.2;
		gbc.weightx=1;
		gbc.fill=GridBagConstraints.HORIZONTAL;
		gbc.gridwidth=3;
		gbc.insets=insets;
		getContentPane().add(subtitleListScrollPane, gbc);
		
		final JList subtitleListBox = new JList();
		final DefaultListModel subtitleListModel=new DefaultListModel();
		subtitleListBox.setModel(subtitleListModel);
		subtitleListBox.setPrototypeCellValue(LIST_BOX_PROTOTYPE);
		subtitleListBox.setVisibleRowCount(4);
		subtitleListBox.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		subtitleListScrollPane.setViewportView(subtitleListBox);

		
		final JTextArea saveAsTextArea = new JTextArea();
		saveAsTextArea.setEditable(false);
		
		/*final JCheckBox force24FpsCheckBox=new JCheckBox("Pulldown (24000/1001 fps)", true);
		force24FpsCheckBox.setEnabled(false);*/
		
		/*gbc = new GridBagConstraints();
		gbc.gridy=++y;
		gbc.anchor=GridBagConstraints.WEST;
		//gbc.insets=insets;
		getContentPane().add(force24FpsCheckBox,gbc);*/
		
		/*final JCheckBox coreCheckBox=new JCheckBox("Core dts-HD and Dolby TrueHD", true);
		coreCheckBox.setEnabled(false);
		
		gbc = new GridBagConstraints();
		gbc.gridy=++y;
		gbc.anchor=GridBagConstraints.WEST;
		getContentPane().add(coreCheckBox,gbc);*/
		
		final JScrollPane outputFileNameScrollPane = new JScrollPane();
		gbc = new GridBagConstraints();
		gbc.fill=GridBagConstraints.BOTH;
		gbc.gridy=++y;
		gbc.gridwidth=2;
		gbc.weightx=1;
		//gbc.weighty=.1;
		gbc.insets=insets;
		getContentPane().add(outputFileNameScrollPane, gbc);
		outputFileNameScrollPane.setViewportView(saveAsTextArea);
		
		final JButton saveAsButton = new JButton(SAVE_AS_TEXT);
		saveAsButton.setEnabled(false);
		gbc = new GridBagConstraints();
		gbc.anchor=GridBagConstraints.NORTH;
		gbc.gridx=1;
		gbc.gridy=y;
		gbc.gridx=2;
		gbc.insets=insets;
		getContentPane().add(saveAsButton, gbc);
		
		final JProgressBar progressBar=new JProgressBar();
		progressBar.setStringPainted(true);
		progressBar.setString("Status");
		gbc = new GridBagConstraints();
//		gbc.weighty=.1;
		gbc.gridy=++y;
		gbc.weightx=1;
		gbc.gridwidth=3;
		gbc.fill=GridBagConstraints.BOTH;
		gbc.insets=insets;
		getContentPane().add(progressBar, gbc);
		
		final JLabel logLabel=new JLabel("Log:");
		gbc = new GridBagConstraints();
		gbc.gridy=++y;
		gbc.anchor=GridBagConstraints.WEST;
		gbc.insets=insets;
		getContentPane().add(logLabel,gbc);
		
		final JScrollPane logScrollPane = new JScrollPane();
		gbc = new GridBagConstraints();
		gbc.gridy=++y;
		gbc.fill=GridBagConstraints.BOTH;
		gbc.gridwidth=3;
		gbc.weightx=1;
		gbc.anchor=GridBagConstraints.SOUTH;
		gbc.weighty=1;
		gbc.insets=insets;
		getContentPane().add(logScrollPane, gbc);
		
		final JTextArea logTextArea = new JTextArea();
		logTextArea.setEditable(false);
		//logTextArea.setRows(1);
		logScrollPane.setViewportView(logTextArea);
		
		final ProgressBarJobListener remuxListener=new ProgressBarJobListener(progressBar);
		
		final class Scanner extends Thread {
			private FileStream fileStream;
			
			private Scanner(FileStream fileStream) {
				this.fileStream=fileStream;
			}
			
			public void run() {
				File files[]=fileStream.getFiles();
				if (StreamConverter.isBluRay(files[0])) {
					converter=new InternalStreamConverter(getPrefs());
				} else {
					Eac3toTsMuxerConverter temp=new Eac3toTsMuxerConverter(getPrefs());
					converter=temp;
					if (files.length==1) {
						files=temp.findFiles(files[0].getAbsolutePath());
					}
				}
				StringBuffer sb=new StringBuffer();
				String sep="";
				for (int i=0;i<files.length;i++) {
					sb.append(sep);
					sb.append(files[i].getAbsolutePath());
					sep="\n";
				}
				inputFileNameTextArea.setText(sb.toString());
				inputFileNameTextArea.setEditable(files.length > 1);
				
				//force24FpsCheckBox.setSelected(!ToNMT.isBluRay(fileStream));
				//force24FpsCheckBox.setEnabled(!ToNMT.isBluRay(fileStream));
				try {
					
					sd=converter.scan(fileStream);
					audioListModel.clear();
					//StreamDescriptor selectedAudio=sd[ToNMT.SD_SELECTED_AUDIO][0];
					for (int i=0;i<sd[StreamConverter.SD_AUDIO].length;i++) {
						audioListModel.addElement(sd[StreamConverter.SD_AUDIO][i]);
						/*if (sd[ToNMT.SD_AUDIO][i]==selectedAudio) {
							audioListBox.setSelectedIndex(i);
						}*/
					}
					audioListBox.repaint();
					
					subtitleListModel.clear();
					for (int i=0;i<sd[StreamConverter.SD_SUBTITLE].length;i++) {
						subtitleListModel.addElement(sd[StreamConverter.SD_SUBTITLE][i]);
					}
					subtitleListBox.repaint();
					//force24FpsCheckBox.setEnabled(!ToNMT.isBluRay(fileStream));
				} catch (Exception e1) {
					logger.error(e1);
					e1.printStackTrace();
					inputFileNameTextArea.setText("");
					saveAsButton.setEnabled(false);
				}
			}
		};
		final ActionListener autoActionListener=new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				try {
					DiskStream diskStreamArray[]=diskFinder.find();
					DiskStream diskStream;
					if (diskStreamArray.length==1) {
						diskStream=diskStreamArray[0];
					} else if (diskStreamArray.length > 0) {
						diskStream = (DiskStream)JOptionPane.showInputDialog(
								ToUI.this,
								"Which source?",
								"Multiple Sources Found",
								JOptionPane.QUESTION_MESSAGE,
								null,
								diskStreamArray,
								diskStreamArray[0]);
					}else {
						logger.warn("No HD Media Found");
						return;
					}
					if (diskStream !=null) {
						
						logger.info("Found: " + diskStream);
						diskStream.setJob(new Job(remuxListener));
						Scanner scanner=new Scanner(diskStream);
						scanner.start();
						saveAsFileChooser.setSelectedFile(new File(getFriendlyName(diskStream)));
					}
				} catch (Exception e1) {
					logger.error(e1);
					e1.printStackTrace();
				}
			}
		};
		autoButton.addActionListener(autoActionListener);
		final ActionListener sourceActionListener=new ActionListener() {
			
			public void actionPerformed(ActionEvent e) {
				int returnVal = inputFileChooser.showOpenDialog(ToUI.this);
				if (returnVal==JFileChooser.APPROVE_OPTION) {
					NameTypeFileFilter fileFilter=(NameTypeFileFilter)inputFileChooser.getFileFilter();
					File[] files=new File[0];
					
					if (PlayListAnalyzer.BLURAY_PLAYLIST_EXT==fileFilter.getFileTypes()[0]) {
						try {
							DiskStream[] diskStreamArray=new PlayListAnalyzer().getTitles(inputFileChooser.getSelectedFiles(), false);
							files=getDiskStreamFiles(diskStreamArray);
						} catch (IOException e1) {
							logger.error(e1);
							e1.printStackTrace();
						}
						
					} else if (XplHelper.HD_HDVD_PLAYLIST_EXT==fileFilter.getFileTypes()[0]) {
						File[] xplFileArray = inputFileChooser.getSelectedFiles();
						List<DiskStream> diskStreamList=new ArrayList<DiskStream>(xplFileArray.length);
						
						for (int i=0;i<xplFileArray.length;i++) {
							try {
								DiskStream diskStream = new XplHelper(xplFileArray[i]).findMainMovie();
								if (diskStream != null) {
									diskStreamList.add(diskStream);
								}
							} catch (Exception e1) {
								logger.error(e1);
								e1.printStackTrace();
							}
						}
						DiskStream[] diskStreamArray=diskStreamList.toArray(new DiskStream[diskStreamList.size()]);
						files=getDiskStreamFiles(diskStreamArray);
					} else {
						files=inputFileChooser.getSelectedFiles();
					}
					if (files.length > 0) {
						inputFileNameTextArea.setText("");
						Scanner scanner=new Scanner(new FileStream(new Job(remuxListener), files));
						scanner.start();
					}
				}
			}
		};
		sourceFileNameButton.addActionListener(sourceActionListener);
		
		final ListSelectionListener audioListSelectionListener=new ListSelectionListener() {
			public void valueChanged(ListSelectionEvent e) {
				Object selected[]=audioListBox.getSelectedValues();
				saveAsButton.setEnabled(selected.length>=1);
			}
		};
		audioListBox.addListSelectionListener(audioListSelectionListener);
		
		final ActionListener preferencesActionListener=new ActionListener() {
			
			public void actionPerformed(ActionEvent e) {
				ToNmtPreferenceDialog prefDialog=getPreferenceDialog();
				prefDialog.setModal(true);
				prefDialog.setVisible(true);
				getPrefs().preferenceCheck();
			}
		};
		preferencesButton.addActionListener(preferencesActionListener);
		
		final ActionListener saveAsActionListener=new ActionListener() {
			
			public void actionPerformed(ActionEvent e) {
				int returnVal = saveAsFileChooser.showSaveDialog(ToUI.this);
				if (returnVal==JFileChooser.APPROVE_OPTION) {
					saveAsButton.setEnabled(false);
					NameTypeFileFilter fileFilter=(NameTypeFileFilter)saveAsFileChooser.getFileFilter();
					File file=saveAsFileChooser.getSelectedFile();

					String fileName=file.getAbsolutePath();
					if (!fileName.toLowerCase().endsWith(fileFilter.getFileTypes()[0])) {
						file=new File(fileName + fileFilter.getFileTypes()[0]);
					}
					boolean created=false;
					try {
						file.delete();
						created=file.createNewFile();
					} catch (IOException ioe) {
						
					}
					if (!created) {
						logger.error("Can not write: " + file.getAbsolutePath());
						return;
					} else {
						file.delete();
					}
					final String outputFileName=file.getAbsolutePath();
					saveAsTextArea.setText(outputFileName);
					Thread thread=new Thread() {
						public void run() {
							try {
								FileStream fileStream=new FileStream(new Job(remuxListener),FileStream.getFiles(inputFileNameTextArea.getText()));
								ArrayList<StreamDescriptor> list=new ArrayList<StreamDescriptor>(1+audioListBox.getSelectedValues().length + subtitleListBox.getSelectedValues().length);
								list.add((VideoStreamDescriptor)sd[StreamConverter.SD_VIDEO][0]);
								appendStreamDescriptors(list,audioListBox.getSelectedValues());
								appendStreamDescriptors(list,subtitleListBox.getSelectedValues());
								logger.info("Selected:");
								for (StreamDescriptor sd : list) {
									logger.info(sd);
								}
								converter.convert(fileStream,outputFileName, list);
								logger.info("Complete");
								converter.postProcess();
							} catch (Exception e1) {
								StringWriter sw=new StringWriter();
								e1.printStackTrace(new PrintWriter(sw));
								logger.error(sw);
								inputFileNameTextArea.setText("");
							}

						}
						private void appendStreamDescriptors(ArrayList<StreamDescriptor> list, Object o[]) {
							for (int i=0;i < o.length;i++) {
								list.add((StreamDescriptor)o[i]);
							}
						}
					};						
					thread.start();
				}
			}
		};
		
		saveAsButton.addActionListener(saveAsActionListener);
		
		final LoggerAppenderI logAppender=new LoggerAppenderI() {
			public void append(String date, String time, String loggerName, Integer level, String message) {
				//int maxLines=PropertiesReader.getInstance().getIntProperty("ui.maxLogLines",1000);
				int maxLines=1000;
				int lines=logTextArea.getLineCount();
				if (lines > maxLines) {
					int firstLine=lines-maxLines;
					try {
						int firstLinePos=logTextArea.getLineEndOffset(firstLine);
						int textLength=logTextArea.getDocument().getLength();
						String logText="... log truncated to " + maxLines + 
							" lines\n" + logTextArea.getText(firstLinePos, textLength-firstLinePos) + 
							time + ' ' + message + "\n";
						logTextArea.setText(logText);
					} catch (BadLocationException e) {
						// Should never happen
						e.printStackTrace();
					}
				} else {
					logTextArea.append(time + ' ' + message + "\n");
				}
				logTextArea.setCaretPosition(logTextArea.getDocument().getLength()-1);
			}
		};
		WindowAdapter windowAdapter=new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				if (converter != null) {
					converter.abort();
				}
				System.exit(0);
			}
		};
		this.addWindowListener(windowAdapter);
		Logger.addAppender(logAppender);
		
		logger.info(getName() + " (Java " + System.getProperty("java.version") + ")");
		getPrefs().preferenceCheck();
	}
	private String getFriendlyName(DiskStream diskStream) {
		String volumeLabel=diskStream.getVolumeLabel();
		
		if (volumeLabel == null) {
			return "";
		} else {
			String driveSeek="Drive (";
			int pos=volumeLabel.indexOf(driveSeek);
			if (pos >= 0) { 
				pos += driveSeek.length() + 3;
				if (pos == volumeLabel.length()) {
					//Escape if it's the default drive XX Drive
					return "";
				} else {
					//Vista format
					volumeLabel=volumeLabel.substring(pos + 1);
				}
			} else {
				//XP Format
				int end=volumeLabel.lastIndexOf('(');
				if (end>0) {
					//Strip off ending drive letter
					volumeLabel=volumeLabel.substring(0, end - 1);
				}
			}
			
			StringBuffer friendlyBuffer=new StringBuffer(volumeLabel.toLowerCase());
			friendlyBuffer.setCharAt(0, Character.toUpperCase(friendlyBuffer.charAt(0)));
			for (int i=1;i<friendlyBuffer.length();i++) {
				char c=friendlyBuffer.charAt(i);
				if (c=='_') {
					friendlyBuffer.setCharAt(i, ' ');
					if (i + 1 < friendlyBuffer.length()) {
						c=friendlyBuffer.charAt(i + 1);
						if (Character.isLowerCase(c)) {
							i++;
							friendlyBuffer.setCharAt(i, Character.toUpperCase(c));
						}
					}
				}
			}
			return friendlyBuffer.toString();
		}
	}
	private File[] getDiskStreamFiles(DiskStream[] diskStreamArray) {
		File files[]=new File[0];
		if (diskStreamArray.length==1) {
			files=diskStreamArray[0].getFiles();
		} else if (diskStreamArray.length>1) {
			DiskStream diskStream = (DiskStream)JOptionPane.showInputDialog(
					ToUI.this,
					"Which source?",
					"Multiple Sources Found",
					JOptionPane.QUESTION_MESSAGE,
					null,
					diskStreamArray,
					diskStreamArray[0]);
			
			if (diskStream != null) {
				files=diskStream.getFiles();
			}
		}
		return files;
	}
}
