package util;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.util.Iterator;
import java.util.Map;

import javax.swing.JFileChooser;

import muxer.MuxManager;
import ui.CliJobListener;

import common.DirectMemoryManager;
import common.FileStream;
import common.Job;
import common.Logger;
import common.MappedBufferForcer;
import common.StreamDescriptor;
import common.mpeg.MpegConstants;
import common.mpeg.MpegUtil;
import common.packet.ClockReferenceI;
import common.ts.ProgramInfo;
import common.ts.ProgramMapTable;
import common.ts.TsConstants;

import demux.ts.M2tsRawDemuxer;
import demux.ts.M2tsRawPacketPart;
import demux.ts.TsInfo;
import demux.ts.TsRawDemuxer;
import demux.ts.TsRawPacketPart;

public class TsMuxerQuickFix implements TsConstants {
	private Logger logger=Logger.getLogger(this.getClass());
	private int AUDIO_PID=0x1100;
	private int AUDIO_PID_MASK=0x1ff0;
	private int PID_MASK=0x1fff;
	
	public boolean fix(FileStream stream) throws IOException {
		File file=stream.getFiles()[0];
		FileChannel inChannel=new RandomAccessFile(file,"rw").getChannel();
		
		MappedByteBuffer buffer=DirectMemoryManager.mapMemory(inChannel, FileChannel.MapMode.READ_ONLY, 0, M2TS_PACKET_SIZE*2);
		int size=0;
		if (buffer.get(0)==TS_SYNC_BYTE && buffer.get(TS_PACKET_SIZE)==TS_SYNC_BYTE) {
			size=TS_PACKET_SIZE;
		} else if (buffer.get(4)==TS_SYNC_BYTE && buffer.get(M2TS_PACKET_SIZE + 4)==TS_SYNC_BYTE) {
			size=M2TS_PACKET_SIZE;
		} else {
			throw new RuntimeException("Could not find TS Sync Byte");
		}
		
		final int maxChunkSize=size*100000;
		buffer=DirectMemoryManager.mapMemory(inChannel, FileChannel.MapMode.READ_WRITE, 0, maxChunkSize);
		long filePos=buffer.capacity();

		//ProgramMapTable pmt=getProgramMapTable(buffer, size);
		ProgramMapTable pmt=getProgramMapTable(stream, size);
		long pcrOffset=0;
		int pcrPid=-1;
		String pcrOffsetString=System.getProperty("pcrOffset");
		if (pcrOffsetString!=null) {
			logger.info("Offseting PCR by: " + pcrOffsetString + "ms");
			pcrOffset=MpegConstants.PTS_FREQ*Long.parseLong(pcrOffsetString)/1000;
			pcrPid=pmt.getPcrPid();
		}
		Map<Integer,StreamDescriptor> streamDescriptorMap=pmt.getSelectedStreamDescriptorMap();
		//int trueHdPid=0x1100;
		//int trueHdPids[]=getTrueHDPid(pmt);
		int ddpPmtPid=getDDPlusPmtPid(pmt);
		ByteBuffer pmtBuffer=null;
		if (/*trueHdPids==null && */ddpPmtPid==-1) {
			inChannel.close();
			return false;
		} 
		logger.info(streamDescriptorMap);
		/*int trueHdPriorityMask[]=null;
		if (trueHdPids != null) {
			trueHdPriorityMask=new int[16];
			for (int i=0;i<trueHdPids.length;i++) {
				if (trueHdPids[i] > 0) {
					logger.info("Found TrueHD PID: 0x" + Integer.toString(trueHdPids[i], 16));
				}
			}
		}*/
		if (ddpPmtPid > 0) {
			logger.info("Found DD+");
			TsRawPacketPart[] tmp=pmt.filter(true);
			pmtBuffer=tmp[0].getByteBuffer();
		}
		logger.info("Fixing...");
		Job job=stream.getJob();
		job.setSize((int)(file.length()/Job.KB));
		job.setDescription("Fixing: " + file.getAbsolutePath());
		job.reset();
		job.setStatus(Job.STATUS_RUNNING);
		job.setPosition(0);
		buffer.position(0);
		buffer.limit(buffer.capacity());
		//int priorityFlag=0;
		
		TsRawPacketPart packet;
		int pidOffset;
		if (size==TS_PACKET_SIZE) {
			packet=new TsRawPacketPart(size);
			pidOffset=1;
		} else {
			packet=new M2tsRawPacketPart(size, 4);
			pidOffset=5;
		}
		//int thdPacketCount=0;
		//int thdPatchCount=0;
		int pmtCount=0;
		while (true) {
			short flags=buffer.getShort(buffer.position()+pidOffset);
			int pid=flags & PID_MASK;
			if ((pid & AUDIO_PID_MASK)==AUDIO_PID) {
				int pidNibble=pid & 0xf;
				//if (trueHdPids[pidNibble]==pid) {
					//thdPacketCount++;
					buffer.limit(buffer.position()+size);
					ByteBuffer packetBuffer=buffer.slice();
					buffer.limit(buffer.capacity());
					packet.setByteBuffer(packetBuffer);
					
					TsInfo tsInfo=packet.getTsInfo();
					if (tsInfo.pidStart) {
						int pesSubStreamOffset=MpegUtil.getPesSubStreamOffset(packet, packet.getPesHeaderOffset());
						packetBuffer.put(pesSubStreamOffset, (byte)0x72);
						/*if (MpegUtil.getPesSubStream(packet, packet.getPesHeaderOffset())==0x76) {
							trueHdPriorityMask[pidNibble]=0xFF;
						} else {
							trueHdPriorityMask[pidNibble]=0xDF;
						}*/
					}
					int b=packetBuffer.get(pidOffset)&0xff;
					int nb=b & 0xDF;
					//if (b!=nb) {
						//thdPatchCount++;
						packetBuffer.put(pidOffset,(byte)nb);
					//}
				//}
				buffer.position(buffer.position()+size);
			} else if (pid==pcrPid) {
				buffer.limit(buffer.position()+size);
				ByteBuffer packetBuffer=buffer.slice();
				buffer.limit(buffer.capacity());
				packet.setByteBuffer(packetBuffer);
				packet.getTsInfo();//Hack to initialize internal structures
				long pcr=packet.getClockReference(ClockReferenceI.CLOCK_REFERENCE_TYPE_SYSTEM);
				packet.setSystemClockReference(pcr + pcrOffset);
				buffer.position(buffer.position()+size);
			} else if (pid==ddpPmtPid) {
				pmtBuffer.rewind();
				buffer.put(pmtBuffer);
				pmtCount++;
			} else {
				//Reminder: Update the audioPid nibbleCheck false scenerio too
				buffer.position(buffer.position()+size);
			}
			if (buffer.position()==buffer.capacity()) {
				if (filePos < inChannel.size()) {
					job.setPosition((int)(filePos/Job.KB));
					long chunkSize=Math.min(maxChunkSize, inChannel.size()-filePos);
					//logger.info("Read: " + filePos);
					new MappedBufferForcer(buffer);
					buffer=DirectMemoryManager.mapMemory(inChannel, FileChannel.MapMode.READ_WRITE, filePos, chunkSize);
					filePos+=buffer.capacity();
				} else {
					break;
				}
			}
		}
		inChannel.close();
		job.done();
		/*if (trueHdPids != null) {
			logger.info("Patched " + thdPatchCount + " of " + thdPacketCount + " TrueHD packets");
		}*/
		if (ddpPmtPid > 0) {
			logger.info("Patched " + pmtCount + " DD+ PMT packets");
		}
		return true;
	}
	
	/*private ProgramMapTable getProgramMapTable(ByteBuffer buffer, int size) {
		int pmtPid=-1;
		final int capacity=buffer.capacity();
		while (capacity - buffer.position() >= size) {
			buffer.limit(buffer.position()+size);
			ByteBuffer packetBuffer=buffer.slice();
			buffer.position(buffer.limit());
			TsRawPacketPart packet=(size==TS_PACKET_SIZE)?new TsRawPacketPart(packetBuffer):new M2tsRawPacketPart(packetBuffer);
			
			TsInfo tsInfo=packet.getTsInfo();
			int pid=tsInfo.pid;
			if (pid==0x0000) {
				if (pmtPid==-1) {
					ProgramAssociationTable pat=new ProgramAssociationTable(new TableHandler());
					pat.append(packet);
					pat.parse();
					ProgramInfo programInfo[]=pat.getProgramInfo();
					for (int i=0;i < programInfo.length;i++) {
						if (programInfo[i].getProgramNumber()==NETWORK_PROGRAM_ID) {
						} else {
							pmtPid=programInfo[i].getPmtPid();
						}
					}
				}
			} else if (pid==pmtPid) {
				ProgramMapTable pmt=new ProgramMapTable(pmtPid, (short)1);
				pmt.append(packet);
				pmt.parse();
				return pmt;
			}
		}
		return null;
	}*/
	private ProgramMapTable getProgramMapTable(FileStream stream, int size) throws IOException {
		TsRawDemuxer demuxer;
		if (size==M2TS_PACKET_SIZE) {
			demuxer=new M2tsRawDemuxer();
		} else {
			demuxer=new TsRawDemuxer();
		}
		MuxManager muxManager=new MuxManager(demuxer);
		FileStream tmpStream=new FileStream(new Job(), stream.getFiles());
		Map <Integer,StreamDescriptor>streamDescriptorMap= muxManager.getStreamDescriptorMap(tmpStream);
		ProgramMapTable pmt=(ProgramMapTable)demuxer.getTableHandler().getTables(ProgramMapTable.class)[0];
		//Select all streams
		pmt.setSelectedStreamDescriptorMap(streamDescriptorMap);
		return pmt;
	}
	
	/*private int[] getTrueHDPid(ProgramMapTable pmt) {
		ProgramInfo programInfo=pmt.getProgramInfo();
		Map<Integer,StreamDescriptor> streamDescriptorMap=programInfo.getStreamDescriptorMap();
		int streamIds[]=new int[16];
		boolean found=false;
		
		for (Iterator<StreamDescriptor> i=streamDescriptorMap.values().iterator();i.hasNext();) {
			StreamDescriptor sd=i.next();
			if (sd.getTsType()==MpegConstants.TS_STREAM_TYPE_TRUEHD_AUDIO) {
				int streamId=sd.getStreamId();
				if ((streamId & AUDIO_PID_MASK) == AUDIO_PID) {
					found=true;
					streamIds[streamId & 0xf]=streamId;
				} else {
					logger.error("Expected TrueHD Stream to be PID 0x110X, but was: 0x" + Integer.toString(streamId,16));
				}
			}
		}
		return found?streamIds:null;
	}*/
	/**
	 * Returns the PMT PID if the PMT contains DD+
	 * @param pmt
	 * @return
	 */
	private int getDDPlusPmtPid(ProgramMapTable pmt) {
		ProgramInfo programInfo=pmt.getProgramInfo();
		//This is kind of a hack.  The selected stream contains the data determine from the stream
		//The is set as part of demuxing (the hack)
		Map<Integer,StreamDescriptor> streamDescriptorMap=pmt.getSelectedStreamDescriptorMap();
		for (Iterator<StreamDescriptor> i=streamDescriptorMap.values().iterator();i.hasNext();) {
			StreamDescriptor sd=i.next();
			if (sd.getTsType()==MpegConstants.TS_STREAM_TYPE_EAC3_AUDIO) {
				return programInfo.getPmtPid();
			}
		} 
		return -1;
	}
	public static void main(String args[]) {
		try {
			File file;
			if (args.length==0) {
				JFileChooser pathChooser=new JFileChooser();
				pathChooser.setMultiSelectionEnabled(false);
				pathChooser.setDialogType(JFileChooser.OPEN_DIALOG);
				int returnVal = pathChooser.showOpenDialog(null);
				if (returnVal==JFileChooser.APPROVE_OPTION) {
					file=pathChooser.getSelectedFile();
				} else {
					System.exit(-1);
					return;
				}
			} else {
				file=new File(args[0]);
			}
			Job job=new Job();
			job.addJobListener(new CliJobListener());
			TsMuxerQuickFix fixer=new TsMuxerQuickFix();
			boolean found=fixer.fix(new FileStream(job, new File[]{file}));
			if (!found) {
				fixer.logger.error("Neither TrueHD nor DD+ Found");
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
}
