package util;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.WritableByteChannel;

import ui.CliJobListener;

import common.BitStream;
import common.BufferedWritableByteChannel;
import common.DirectMemoryManager;
import common.Job;
import common.Util;
import common.mpeg.MpegUtil;
import common.vc1.Vc1Constants;
import common.vc1.Vc1Deinterlacer;

public class Vc1ConvJ extends Vc1Deinterlacer implements Vc1Constants {
	private static final int READ_SIZE=20*1024*1024;
	private WritableByteChannel out;
	
	public static void main(String args[]) {
		long start=System.currentTimeMillis();
		try {
			System.out.println("Vc1ConvJ - HD DVD to BluRay compliant VC1 processor");
			if (args.length!=2) {
				System.out.println("Usage:");
				System.out.println("Vc1ConvJ inVC1File outVC1File");
				System.exit(-1);
				//args=new String[]{"f:/Bourne_fd55.vc1","C:/my_Bourne2_fd55.vc1"};
			}
			Vc1ConvJ test=new Vc1ConvJ(args[1]);
			System.out.println("Processing " + args[0]);
			test.parse(args[0]);
			System.out.println();
			System.out.println("Completed in " + Util.timeToString(System.currentTimeMillis()-start));
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	public Vc1ConvJ(String fileName) throws IOException {
		//out=new FileOutputStream(new File(fileName)).getChannel();
		out=new BufferedWritableByteChannel(new File(fileName),10*1024*1024);
	}
	public void parse(String fileName) throws IOException {
		long filePosition=0;
		FileChannel fc=new FileInputStream(fileName).getChannel();
		Job job=new Job(new CliJobListener());
		job.setSize((int)(fc.size()/Job.KB));
		ByteBuffer byteBuffer=ByteBuffer.allocate(0);//Dummy
		boolean lastPacket=false;
		BitStream bitStream=new BitStream(2*1024*1024);
		BitStream inBitStream=new BitStream(new byte[0]);
		byte type=-1;
		int typeStart=-1;
		while (!lastPacket) {
			filePosition+=byteBuffer.position();
			job.setPosition((int)(filePosition/Job.KB));
			int readSize=(int)Math.min(READ_SIZE, fc.size() - filePosition);
			lastPacket=readSize!=READ_SIZE;
			byteBuffer=DirectMemoryManager.mapMemory(fc, FileChannel.MapMode.READ_ONLY, filePosition, readSize);
			type=-1;
			typeStart=-1;
			for (int start=0;start>=0;) {
				start=MpegUtil.findNextStartCode(byteBuffer, start);
				if (start>=0) {
					byte nextType=byteBuffer.get(start+3);
					if (nextType==PICTURE_START || nextType==ENTRY_POINT_START || nextType==SEQUENCE_START) {
						final int len=start-typeStart;
						if (len*105/100 > bitStream.remaining()) {
							bitStream.write(out);
							bitStream.limit(0);
							bitStream.setPosition(0);
							//bitStream=new Vc1BitStream(2*1024*1024);
						}
						if (type==PICTURE_START) {
							inBitStream.setByteBuffer(byteBuffer);
							inBitStream.setPosition(typeStart);
							copyPicture(start, inBitStream, bitStream);
						} else if (type==ENTRY_POINT_START) {
							copyEntryPoint(typeStart, start, byteBuffer, bitStream);
						} else if (type==SEQUENCE_START) {
							copySequence(typeStart, start, byteBuffer, bitStream);
						} 
						type=nextType;
						typeStart=start;
					}
					start+=4;
				}
			}
			byteBuffer.position(typeStart);
		}
		//Should always end on a picture
		inBitStream.setByteBuffer(byteBuffer);
		inBitStream.setPosition(typeStart);
		copyPicture(byteBuffer.capacity(), inBitStream, bitStream);
		bitStream.write(out);
		out.close();
		fc.close();
	}

}
