package web;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import common.Logger;
import common.ThreadPool;


public abstract class WebServer implements Runnable {
	private int port;
	protected List servlets=new ArrayList();
	protected SecurityInterceptorI securityInterceptor;
	protected Logger logger;
	
	public WebServer(int port) {
		this.port=port;
	}
	
	public void run() {
		ServerSocket server;
		try {
			server=new ServerSocket(port);
			
		} catch (IOException ioe) {
			System.out.println(ioe);
			ioe.printStackTrace();
			return;
		}
		ThreadPool threadPool=new ThreadPool();
		logger.info("HD Transcode on http://localhost:" + port +"/");
		while (true) {
			try {
				ServerConnection serverConnection=new ServerConnection();
				Socket socket = server.accept();
				serverConnection.init(socket, this);
				threadPool.runTask(serverConnection);
			} catch (IOException ioe) {
				logger.debug(ioe);
				ioe.printStackTrace();
			}
		}
	}
	
	protected MinimalServlet getServlet(String path) {
		for (Iterator i=servlets.iterator();i.hasNext();) {
			ServletMap servletMap=(ServletMap)i.next();
			if (path.startsWith(servletMap.getPath())) {
				return servletMap.getMinimalServlet();
			}
		}
		logger.warn("No servlet found for path: " + path);
		return null;
	}
	
	public SecurityInterceptorI getSecurityInterceptor() {
		return securityInterceptor;
	}
}
